from .weather_station_test_case import WeatherStationTestCase
from .collection_test_case import CollectionTestCase
from .file_storage_test_case import FileStorageTestCase
from .raster_band_test_case import RasterBandTestCase
from .wind_rose_test_case import WindRoseTestCase
from .kriger_test_case import KrigerTestCase



import logging
import os
import numpy as np
import numpy.ma as ma
from osgeo import gdal, ogr
from .file_storage import FileStorage

logger = logging.getLogger(__name__)

class RasterBand:

  def __init__(self):
    self._tif = None
    self._band = None
    self._nodata = None
    self._cols = None
    self._rows = None
    self._x_origin = None
    self._y_origin = None
    self._pixel_w = None
    self._pixel_h = None

  def load_url(self, file_url, md5_hash=None):
    logger.debug("opening %s" % file_url)
    fs = FileStorage()
    fs.set_file_url(file_url)
    fs.download()
    self.loadf(fs.get_full_path_to_file())
  
  def loadf(self, gtif):
    logger.debug("opening %s" % gtif)
    if not os.path.exists(gtif):
      raise ValueError(f"File does not exist: {gtif}")
    self._tif = gdal.Open(gtif)

  def band_count(self):
    return self._tif.RasterCount

  def load_band(self, band_id=1):
    # make sure the band exists
    if band_id > self.band_count():
      raise ValueError(f"Invalid band id: {band_id}")
    self._band = self._tif.GetRasterBand(band_id)
    self._nodata = self._band.GetNoDataValue()
    logger.debug("nodata value = %s" % self._nodata)
    self._cols = self._tif.RasterXSize
    self._rows = self._tif.RasterYSize
    # geotransform is (top left x, w-e pixel resolution, rotation, top left y, rotation, n-s pixel resolution)
    transform = self._tif.GetGeoTransform()
    self._x_origin = transform[0] # top left x
    self._y_origin = transform[3] # top left y
    self._pixel_w = transform[1] # w-e pixel resolution
    self._pixel_h = -transform[5] # n-s pixel resolution, negative because y increases down
    self._we_res = transform[1] # w-e pixel resolution
    self._ns_res = transform[5] # n-s pixel resolution

  def shape(self):
    self._throw_except_if_band_not_loaded()
    return (self._cols, self._rows)

  def size(self):
    self._throw_except_if_band_not_loaded()
    return self._cols * self._rows

  def _throw_except_if_band_not_loaded(self):
    if self._band is None:
      raise ValueError("No band loaded")

  # returns the centroid for a given box, where the point
  # is the UL point
  def _centroid(self, x, y):
    center_x = x + (self._pixel_w/2)
    center_y = y + (self._pixel_h/2)
    return (center_x, center_y)

  def rows(self, return_centroids=False):
    for i in range(self._rows):
      y = self._y_origin - (i * self._pixel_h)
      x = self._x_origin
      if return_centroids is True:
        coords = [self._centroid(x + (j * self._pixel_w), y) \
          for j in range(self._cols)]
      else:
        coords = [(x + (j * self._pixel_w), y) for j in range(self._cols)]
      data = self._band.ReadAsArray(0, i, self._cols, 1)[0]
      data = [d if d != self._nodata else None for d in data]
      yield list(zip(coords, data))

  def get_bbox_polygon(self):
    self._throw_except_if_band_not_loaded()
    bbox = self.get_bbox()
    ring = ogr.Geometry(ogr.wkbLinearRing)
    ring.AddPoint(bbox[0][0], bbox[0][1])
    ring.AddPoint(bbox[1][0], bbox[0][1])
    ring.AddPoint(bbox[1][0], bbox[1][1])
    ring.AddPoint(bbox[0][0], bbox[1][1])
    ring.AddPoint(bbox[0][0], bbox[0][1])
    poly = ogr.Geometry(ogr.wkbPolygon)
    poly.AddGeometry(ring)
    return poly

  # return tupble of (ul, lr) coordinates
  def get_bbox(self):
    self._throw_except_if_band_not_loaded()
    ul = (self._x_origin, self._y_origin)
    lr = (self._x_origin + self._cols * self._pixel_w,
      self._y_origin - self._rows * self._pixel_h)
    return (ul, lr)

  # get the centerpoint of the raster in the coordinate system
  # the raster is in
  def get_center(self):
    self._throw_except_if_band_not_loaded()
    return (self._x_origin + (self._cols * self._pixel_w / 2.0),
            self._y_origin - (self._rows * self._pixel_h / 2.0))

  def get_nodata(self):
    return self._nodata

  def set_nodata(self, nodata):
    self._nodata = nodata

  # retrieve a single value at a location
  # the x,y values are in the coordinate system of the raster
  def get_value(self, x, y):
    self._throw_except_if_band_not_loaded()

    point = ogr.Geometry(ogr.wkbPoint)
    point.AddPoint(x, y)

    # get the bbox and determine if the point is inside of it
    # by using gdal functions + intersection
    bbox_polygon = self.get_bbox_polygon()
    if not bbox_polygon.Contains(point):
      raise ValueError(f"Point outside raster: {x}, {y}")

    # now we need to figure out which row and column we are in
    # make sure to consider resolution and negative y axis
    col = int(np.floor(np.abs(self._y_origin - y) / self._pixel_h)) 
    row = int(np.floor(np.abs(x - self._x_origin) / self._pixel_w))

    data = self._band.ReadAsArray(row, col, 1, 1).tolist()[0][0]

    if data == self._nodata:
      return None

    return data



# file_storage.py
# this is a file storage class which is used to download
# a file from a URL and store it locally
# it also has a method to check the md5 hash of the file
# and if it doesn't match, it will raise an exception
# author: Tom Hayden

import json
import hashlib
import os
import platform
import logging
import datetime
import requests

logger = logging.getLogger(__name__)

class FileStorage:

  def __init__(self):
    self.set_cache_dir()
    self.file_url = None
    self.expected_md5_hash = None
    self.basename = None

  def get_cache_dir(self):
    return self.cache_dir

  def set_file_url(self, file_url):
    self.file_url = file_url

    # figure out the basename from the http url
    self.basename = os.path.basename(self.file_url)

  def set_cache_dir(self):
    if platform.system() == 'Windows':
      self.cache_dir = os.getenv('LOCALAPPDATA')
    else:
      self.cache_dir = os.path.expanduser('~/.cache')
    self.cache_dir = os.path.join(self.cache_dir, 'wx_logs')
    if not os.path.exists(self.cache_dir):
      os.makedirs(self.cache_dir)
 
  def get_full_path_to_file(self):
    return os.path.join(self.cache_dir, self.basename)

  def set_expected_md5_hash(self, md5_hash):
    self.expected_md5_hash = md5_hash

  # return ONLY the basename of the file
  def get_file_name(self):
    return self.basename

  def get_file_size(self):
    return os.path.getsize(self.get_full_path_to_file())

  # md5 hash of the file
  def get_md5_hash(self):
    file_path = self.get_full_path_to_file()
    with open(file_path, 'rb') as f:
      return hashlib.md5(f.read()).hexdigest()

  def delete_file(self):
    file_name = self.get_file_name()
    if os.path.exists(file_name):
      os.remove(file_name)

  # this the main call to download the file, but we should also 
  # make sure to see if it's already there and if so, check the md5 hash
  # and in that case just return bc we are ok
  def download(self):
    if self.file_url is None:
      raise ValueError("No file URL set. call set_file_url()")
    file_name = self.get_full_path_to_file()
    if os.path.exists(file_name):
      if self.expected_md5_hash is not None:
        actual_md5_hash = self.get_md5_hash()
        if actual_md5_hash != self.expected_md5_hash:
          raise ValueError(f"MD5 hash mismatch for file: {self.file_url}")
      return
    logger.info(f"Downloading file: {self.file_url}")
    response = requests.get(self.file_url)
    if response.status_code != 200:
      raise ValueError(f"Invalid response code: {response.status_code}")

    with open(file_name, 'wb') as f:
      f.write(response.content)
    logger.info(f"Downloaded file: {self.file_url}")

    if self.expected_md5_hash is not None:
      actual_md5_hash = self.get_md5_hash()
      if actual_md5_hash != self.expected_md5_hash:
        raise ValueError(f"MD5 hash mismatch for file: {self.file_url}")



import json
import os
import joblib
import logging
import datetime
from .weather_station import WeatherStation
from .kriger import Kriger
from .raster_band import RasterBand
from .file_storage import FileStorage

logger = logging.getLogger(__name__)

MISMATCH_THRESH_M = 25

class Collection:

  def __init__(self):
    self._stations = {}
    self._dem_band = None

  def add_station(self, station):
    if not isinstance(station, WeatherStation):
      logger.error("Invalid station type")
      return
    station_id = station.get_station_id()
    if station_id is None:
      raise Exception("Cannot add station without an ID!")
    station_id = str(station_id)
    self._stations[station_id] = station

  # downloads the DEM data
  def _get_dem_band(self, file_location, file_md5):
    s = FileStorage()
    s.set_file_url(file_location)
    s.set_expected_md5_hash(file_md5)
    s.download()
    self._dem_band = RasterBand()
    self._dem_band.loadf(s.get_full_path_to_file())
    self._dem_band.load_band(1)

  # append elevation will combine DEM elevation rasters
  # with the station data, to make sure we can set the right
  # elevation on all the points
  def append_elevations_from_dem(self, dem_s3_path, dem_md5, none_to_zero=True):
    if self._dem_band is None:
      logger.warning("DEM Band needs to download. Please wait.")
      self._get_dem_band(dem_s3_path, dem_md5)
    for (station_id, station) in self._stations.items():
      location = station.get_location()
      latitude = location['latitude']
      longitude = location['longitude']
      current_elevation = station.get_elevation()
      matched_val = self._dem_band.get_value(longitude, latitude)

      # if matched_val is None, that usually means ocean, so elev = 0
      if none_to_zero and matched_val is None:
        matched_val = 0.0

      if current_elevation is not None and matched_val is not None:
        if abs(current_elevation - matched_val) > MISMATCH_THRESH_M:
          station.set_elevation(matched_val)
      else:
        station.set_elevation(matched_val)

  # use joblib to save to file, use gzip
  def save(self, filename):
    assert filename is not None, "Filename cannot be None"
    joblib.dump(self, filename, compress=3)

  # load from file
  @staticmethod
  def load(filename):
    assert filename is not None, "Filename cannot be None"
    return joblib.load(filename)

  # this will take a single variable on all stations
  # and krige it over some period of time
  def krige_single_variable(self, variable_name):
    pass

  def new_station(self, station_type, station_id):
    station_id = str(station_id)
    s = WeatherStation(station_type)
    s.set_station_id(station_id)
    self.add_station(s)
    return s

  def num_stations(self):
    return len(self._stations)

  def get_station_by_id(self, station_id):
    station_id = str(station_id)
    return self._stations.get(station_id)

  def stations(self):
    for (station_id, station) in self._stations.items():
      yield station



from .weather_station import WeatherStation
from .collection import Collection
from .file_storage import FileStorage
from .raster_band import RasterBand
from .wind_rose import WindRose
from .kriger import Kriger



from setuptools import setup, find_packages

setup(
  name='wx_logs',
  version='0.4.9',
  author='Tom Hayden',
  author_email='thayden@gmail.com',
  packages=find_packages(exclude=['tests', 'tests.*']),
  long_description="A weather logging library, useful for processing and analyzing weather data.",
  include_package_data=True,
  install_requires=['dateparser', 'numpy', 'pytz', 'gdal', 'gstools', 'windrose', 'matplotlib', 'joblib'],
  entry_points={
    'console_scripts': [],
  },
  classifiers=[
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3',
  ],
  python_requires='>=3.6',
)

